@@ -19,26 +19,25 @@
 # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 # SOFTWARE.
 
+from __future__ import annotations
+
 import copy
 from abc import ABCMeta, abstractmethod
-from typing import Any, List, Tuple  # noqa: F401
+from typing import TYPE_CHECKING
 
-from .. import configurable
-from libqtile.command_object import CommandObject
+from libqtile import configurable
+from libqtile.command.base import CommandObject
+
+if TYPE_CHECKING:
+    from typing import Any
+
+    from libqtile.command.base import ItemT
 
 
 class Layout(CommandObject, configurable.Configurable, metaclass=ABCMeta):
     """This class defines the API that should be exposed by all layouts"""
-    @classmethod
-    def _name(cls):
-        return cls.__class__.__name__.lower()
 
-    defaults = [(
-        "name",
-        None,
-        "The name of this layout"
-        " (usually the class' name in lowercase, e.g. 'max')"
-    )]  # type: List[Tuple[str, Any, str]]
+    defaults = []  # type: list[tuple[str, Any, str]]
 
     def __init__(self, **config):
         # name is a little odd; we can't resolve it until the class is defined
@@ -51,10 +50,10 @@ class Layout(CommandObject, configurable.Configurable, metaclass=ABCMeta):
         configurable.Configurable.__init__(self, **config)
         self.add_defaults(Layout.defaults)
 
-    def layout(self, windows, screen):
+    def layout(self, windows, screen_rect):
         assert windows, "let's eliminate unnecessary calls"
         for i in windows:
-            self.configure(i, screen)
+            self.configure(i, screen_rect)
 
     def finalize(self):
         pass
@@ -74,11 +73,12 @@ class Layout(CommandObject, configurable.Configurable, metaclass=ABCMeta):
         c.group = group
         return c
 
-    def _items(self, name):
-        if name == "screen":
-            return (True, None)
+    def _items(self, name: str) -> ItemT:
+        if name == "screen" and self.group.screen is not None:
+            return True, []
         elif name == "group":
-            return (True, None)
+            return True, []
+        return None
 
     def _select(self, name, sel):
         if name == "screen":
@@ -86,7 +86,7 @@ class Layout(CommandObject, configurable.Configurable, metaclass=ABCMeta):
         elif name == "group":
             return self.group
 
-    def show(self, screen):
+    def show(self, screen_rect):
         """Called when layout is being shown"""
         pass
 
@@ -104,10 +104,7 @@ class Layout(CommandObject, configurable.Configurable, metaclass=ABCMeta):
 
     def info(self):
         """Returns a dictionary of layout information"""
-        return dict(
-            name=self.name,
-            group=self.group.name if self.group else None
-        )
+        return dict(name=self.name, group=self.group.name if self.group else None)
 
     def cmd_info(self):
         """Return a dictionary of info for this object"""
@@ -136,7 +133,7 @@ class Layout(CommandObject, configurable.Configurable, metaclass=ABCMeta):
         pass
 
     @abstractmethod
-    def configure(self, client, screen):
+    def configure(self, client, screen_rect):
         """Configure the layout
 
         This method should:
@@ -216,145 +213,6 @@ class Layout(CommandObject, configurable.Configurable, metaclass=ABCMeta):
         pass
 
 
-class SingleWindow(Layout):
-    """Base for layouts with single visible window"""
-
-    def __init__(self, **config):
-        Layout.__init__(self, **config)
-
-    @abstractmethod
-    def _get_window(self):
-        """Should return either visible window or None"""
-        pass
-
-    def configure(self, win, screen):
-        if win is self._get_window():
-            win.place(
-                screen.x, screen.y,
-                screen.width, screen.height,
-                0,
-                None,
-            )
-            win.unhide()
-        else:
-            win.hide()
-
-    def remove(self, win):
-        cli = self.clients.pop(0)
-        if cli == win:
-            return self.clients[0]
-
-
-class Delegate(Layout):
-    """Base for all delegation layouts"""
-
-    def __init__(self, **config):
-        self.layouts = {}
-        Layout.__init__(self, **config)
-
-    def clone(self, group):
-        c = Layout.clone(self, group)
-        c.layouts = {}
-        return c
-
-    @abstractmethod
-    def _get_layouts(self):
-        """Returns all children layouts"""
-        pass
-
-    @abstractmethod
-    def _get_active_layout(self):
-        """Returns layout to which delegate commands to"""
-        pass
-
-    def delegate_layout(self, windows, mapping):
-        """Delegates layouting actual windows
-
-        Parameterer
-        ===========
-        windows:
-            windows to layout
-        mapping:
-            mapping from layout to ScreenRect for each layout
-        """
-        grouped = {}
-        for w in windows:
-            lay = self.layouts[w]
-            if lay in grouped:
-                grouped[lay].append(w)
-            else:
-                grouped[lay] = [w]
-        for lay, wins in grouped.items():
-            lay.layout(wins, mapping[lay])
-
-    def remove(self, win):
-        lay = self.layouts.pop(win)
-        focus = lay.remove(win)
-        if not focus:
-            layouts = self._get_layouts()
-            idx = layouts.index(lay)
-            while idx < len(layouts) - 1 and not focus:
-                idx += 1
-                focus = layouts[idx].focus_first()
-        return focus
-
-    def focus_first(self):
-        layouts = self._get_layouts()
-        for lay in layouts:
-            win = lay.focus_first()
-            if win:
-                return win
-
-    def focus_last(self):
-        layouts = self._get_layouts()
-        for lay in reversed(layouts):
-            win = lay.focus_last()
-            if win:
-                return win
-
-    def focus_next(self, win):
-        layouts = self._get_layouts()
-        cur = self.layouts[win]
-        focus = cur.focus_next(win)
-        if not focus:
-            idx = layouts.index(cur)
-            while idx < len(layouts) - 1 and not focus:
-                idx += 1
-                focus = layouts[idx].focus_first()
-        return focus
-
-    def focus_previous(self, win):
-        layouts = self._get_layouts()
-        cur = self.layouts[win]
-        focus = cur.focus_previous(win)
-        if not focus:
-            idx = layouts.index(cur)
-            while idx > 0 and not focus:
-                idx -= 1
-                focus = layouts[idx].focus_last()
-        return focus
-
-    def __getattr__(self, name):
-        """Delegate unimplemented command calls to active layout.
-
-        For ``cmd_``-methods that don't exist on the Delegate subclass, this
-        looks for an implementation on the active layout.
-        """
-        if name.startswith('cmd_'):
-            return getattr(self._get_active_layout(), name)
-        return super().__getattr__(name)
-
-    @property
-    def commands(self):
-        return self._get_active_layout().commands
-
-    def info(self):
-        d = Layout.info(self)
-        for layout in self._get_layouts():
-            d[layout.name] = layout.info()
-        return d
-
-
 class _ClientList:
     """
     ClientList maintains a list of clients and a current client.
@@ -427,27 +285,37 @@ class _ClientList:
         """
         Returns the client previous to win in collection.
         """
-        try:
-            idx = self.index(win)
-        except IndexError:
-            return None
-        else:
-            if idx > 0:
-                return self[idx - 1]
+        idx = self.index(win)
+        if idx > 0:
+            return self[idx - 1]
+        return None
 
-    def add(self, client, offset_to_current=0):
+    def add(self, client, offset_to_current=0, client_position=None):
         """
         Insert the given client into collection at position of the current.
 
         Use parameter 'offset_to_current' to specify where the client shall be
         inserted. Defaults to zero, which means at position of current client.
         Positive values are after the client.
+
+        Use parameter 'client_position' to insert the given client at 4 specific
+        positions: top, bottom, after_current, before_current.
         """
-        pos = max(0, self._current_idx + offset_to_current)
-        if pos < len(self.clients):
-            self.clients.insert(pos, client)
+        if client_position is not None:
+            if client_position == "after_current":
+                return self.add(client, offset_to_current=1)
+            elif client_position == "before_current":
+                return self.add(client, offset_to_current=0)
+            elif client_position == "top":
+                self.append_head(client)
+            else:  # ie client_position == "bottom"
+                self.append(client)
         else:
-            self.clients.append(client)
+            pos = max(0, self._current_idx + offset_to_current)
+            if pos < len(self.clients):
+                self.clients.insert(pos, client)
+            else:
+                self.clients.append(client)
         self.current_client = client
 
     def append_head(self, client):
@@ -543,8 +411,7 @@ class _ClientList:
         """
         pos = max(0, self.current_index + offset_to_current)
         if pos < len(self.clients):
-            self.clients = (self.clients[:pos:] + other.clients +
-                            self.clients[pos::])
+            self.clients = self.clients[:pos:] + other.clients + self.clients[pos::]
         else:
             self.clients.extend(other.clients)
 
@@ -572,7 +439,8 @@ class _ClientList:
     def __str__(self):
         curr = self.current_client
         return "_WindowCollection: " + ", ".join(
-            [('[%s]' if c == curr else '%s') % c.name for c in self.clients])
+            [("[%s]" if c == curr else "%s") % c.name for c in self.clients]
+        )
 
     def info(self):
         return dict(
@@ -627,12 +495,15 @@ class _SimpleLayoutBase(Layout):
         client = self.focus_next(self.clients.current_client) or self.focus_first()
         self.group.focus(client, True)
 
-    def add(self, client, offset_to_current=0):
-        return self.clients.add(client, offset_to_current)
+    def add(self, client, offset_to_current=0, client_position=None):
+        return self.clients.add(client, offset_to_current, client_position)
 
     def remove(self, client):
         return self.clients.remove(client)
 
+    def get_windows(self):
+        return self.clients.clients
+
     def info(self):
         d = Layout.info(self)
         d.update(self.clients.info())
